#!/usr/bin/env python

# ~ import niquests
import base64

import requests
from requests import Session, exceptions
from requests.adapters import HTTPAdapter
from requests.packages.urllib3.util.ssl_ import create_urllib3_context
from urllib3 import PoolManager
from lxml import etree

from settings import TIMEOUT, log


requests.packages.urllib3.disable_warnings()


class CipherAdapter(HTTPAdapter):
    def init_poolmanager(self, connections, maxsize, block=False):
        ctx = create_urllib3_context(ciphers=":HIGH:!DH:!aNULL")
        self.poolmanager = PoolManager(
            num_pools=connections, maxsize=maxsize, block=block, ssl_context=ctx)


class SATPortal(object):
    BROWSER = 'Mozilla/5.0 (X11; Linux x86_64; rv:55.0) Gecko/20100101 Firefox/55.0'
    HOST = 'cfdiau.sat.gob.mx'
    REFERER = 'https://cfdiau.sat.gob.mx/nidp/app/login?id=SATUPCFDiCon&sid=0&option=credential&sid=0'
    URL = {
        'MAIN': 'https://portalcfdi.facturaelectronica.sat.gob.mx/',
        'PORTAL': 'https://portalcfdi.facturaelectronica.sat.gob.mx/',
    }
    URL['LOGOUT'] = f"{URL['PORTAL']}logout.aspx?salir=y"
    URL['CONSULTA'] = f"{URL['PORTAL']}Consulta.aspx"

    def __init__(self):
        # ~ self._s = niquests.Session()
        self._s = Session()
        self._s.mount('https://', CipherAdapter())

    def _add_headers(self, headers):
        for k, v in headers.items():
            self._s.headers[k] = v
        return

    def _get(self, url):
        # ~ log.debug('URL: {}'.format(url))
        try:
            result = self._s.get(url, timeout=TIMEOUT)
            if result.status_code != 200:
                msg = f'Code: {result.status_code}. URL: {url}'
                log.error(msg)
                return ''
        except exceptions.Timeout:
            msg = 'Tiempo de espera agotado'
            self.not_network = True
            log.error(msg)
            self.error = msg
            return ''
        except exceptions.ConnectionError:
            msg = 'Revisa la conexión a Internet'
            self.not_network = True
            log.error(msg)
            self.error = msg
            return ''

        return result.text

    def _post(self, url, data={}):
        # ~ log.debug('URL: {}'.format(url))
        try:
            result = self._s.post(url, data=data, timeout=TIMEOUT)
            if result.status_code != 200:
                msg = f'Code: {result.status_code}. URL: {url}'
                log.error(msg)
                return ''
        except exceptions.Timeout:
            msg = 'Tiempo de espera agotado'
            self.not_network = True
            log.error(msg)
            self.error = msg
            return ''
        except exceptions.ConnectionError:
            msg = 'Revisa la conexión a Internet'
            self.not_network = True
            log.error(msg)
            self.error = msg
            return ''

        return result.text

    def _read_form(self, html, login=False):
        values = {}
        parser = etree.HTMLParser()
        tree = etree.fromstring(html, parser)

        if login:
            for node in tree.xpath('//input'):
                try:
                    values[node.attrib['id']] = node.attrib['value']
                except:
                    pass
        else:
            for node in tree.xpath('//input | //select'):
                a = node.attrib
                if a.get('type', '') and a['type'] == 'hidden':
                    if 'name' in a and 'value' in a:
                        values[a['name']] = a['value']

        return values

    def _get_headers(self, host, referer, ajax=False):
        acept = 'text/html,application/xhtml+xml,application/xml;q=0.9,*/*;q=0.8'
        headers = {
            'Accept': acept,
            'Accept-Encoding': 'gzip, deflate, br',
            'Accept-Language': 'en-US,en;q=0.5',
            'Connection': 'keep-alive',
            'DNT': '1',
            'Host': host,
            'Referer': referer,
            'Upgrade-Insecure-Requests': '1',
            'User-Agent': self.BROWSER,
            'Content-Type': 'application/x-www-form-urlencoded',
        }
        if ajax:
            headers.update({
                'Cache-Control': 'no-cache',
                'X-MicrosoftAjax': 'Delta=true',
                'x-requested-with': 'XMLHttpRequest',
                'Pragma': 'no-cache',
            })
        return headers

    def _get_token(self, firma, co):
        co = base64.b64encode(co.encode('utf-8')).decode('utf-8')
        data = f'{co}#{firma}'.encode('utf-8')
        token = base64.b64encode(data).decode('utf-8')
        return token

    def _sign(self, values, fiel):
        fert = f'{fiel.cer.not_after:%y%m%d%H%M%SZ}'
        co = f"{values['tokenuuid']}|{fiel.cer.rfc}|{fiel.cer.serial_number2}"
        firma = base64.b64encode(fiel.sign(co).encode()).decode('utf-8')
        token = self._get_token(firma, co)

        keys = ('credentialsRequired', 'guid', 'ks', 'urlApplet')
        data = {k: values[k] for k in keys}
        data['fert'] = fert
        data['token'] = token
        data['arc'] = ''
        data['placer'] = ''
        data['secuence'] = ''
        data['seeder'] = ''
        data['tan'] = ''
        return data

    def login(self, fiel):
        HOST = 'cfdicontribuyentes.accesscontrol.windows.net'
        REFERER = 'https://cfdiau.sat.gob.mx/nidp/wsfed/ep?id=SATUPCFDiCon&sid=0&option=credential&sid=0'
        URL_LOGIN = 'https://cfdiau.sat.gob.mx/nidp/app/login?id=SATx509Custom&sid=0&option=credential&sid=0'

        result = self._s.get(self.URL['MAIN'])
        url_redirect = result.history[-1].headers['Location']
        self._add_headers({'Host': self.HOST})
        result = self._get(url_redirect)
        headers = {
            'User-Agent': self.BROWSER,
            'Referer': f'{REFERER}{url_redirect}'}
        self._add_headers(headers)
        result = self._post(URL_LOGIN)
        values = self._read_form(result, True)
        data = self._sign(values, fiel)
        headers = self._get_headers(self.HOST, self.REFERER)
        self._add_headers(headers)
        result = self._post(URL_LOGIN, data)

        if not result:
            msg = 'Error al identificarse en el SAT'
            log.error(msg)
            return False

        data = self._read_form(result)

        # Inicio
        result = self._post(self.URL['MAIN'], data)
        data = self._get_post_type_search(result)
        headers = self._get_headers(self.HOST, self.URL['MAIN'])

        # Consulta
        self._add_headers(headers)
        result = self._post(self.URL['CONSULTA'], data)
        msg = 'Se ha identificado en el SAT'
        log.info(msg)
        self.is_connect = True
        return True

    def _get_post_type_search(self, html, emitidas=False):
        tipo_busqueda = 'RdoTipoBusquedaReceptor'
        if emitidas:
            tipo_busqueda = 'RdoTipoBusquedaEmisor'
        sm = 'ctl00$MainContent$UpnlBusqueda|ctl00$MainContent$BtnBusqueda'
        post = self._read_form(html)
        post['ctl00$MainContent$TipoBusqueda'] = tipo_busqueda
        post['__ASYNCPOST'] = 'true'
        post['__EVENTTARGET'] = ''
        post['__EVENTARGUMENT'] = ''
        post['ctl00$ScriptManager1'] = sm
        return post

    def logout(self):
        msg = 'Cerrando sessión en el SAT'
        log.debug(msg)
        result = self._get(self.URL['LOGOUT'])
        self.is_connect = False
        msg = 'Sesión cerrada en el SAT'
        log.info(msg)
        return